

#if defined(__SAM3X8E__)
	#include <include/pio.h>
    #define PROGMEM
    #define pgm_read_byte(addr) (*(const unsigned char *)(addr))
    #define pgm_read_word(addr) (*(const unsigned short *)(addr))
#endif
#ifdef __AVR__
	#include <avr/pgmspace.h>
#endif
#include "pins_arduino.h"
#include "wiring_private.h"
#include "ST7783.h"
#include "pin_magic.h"

#define TFTWIDTH   240    
#define TFTHEIGHT  320


// Constructor for breakout board (configurable LCD control lines).
// Can still use this w/shield, but parameters are ignored.
ST7783::ST7783(
  uint8_t cs, uint8_t cd, uint8_t wr, uint8_t rd, uint8_t reset) :
  Adafruit_GFX(TFTWIDTH, TFTHEIGHT) {

#ifndef USE_ADAFRUIT_SHIELD_PINOUT
  // Convert pin numbers to registers and bitmasks
  _reset     = reset;
  #ifdef __AVR__
    csPort     = portOutputRegister(digitalPinToPort(cs));
    cdPort     = portOutputRegister(digitalPinToPort(cd));
    wrPort     = portOutputRegister(digitalPinToPort(wr));
    rdPort     = portOutputRegister(digitalPinToPort(rd));
  #endif
  #if defined(__SAM3X8E__)
    csPort     = digitalPinToPort(cs);
    cdPort     = digitalPinToPort(cd);
    wrPort     = digitalPinToPort(wr);
    rdPort     = digitalPinToPort(rd);
  #endif
  csPinSet   = digitalPinToBitMask(cs);
  cdPinSet   = digitalPinToBitMask(cd);
  wrPinSet   = digitalPinToBitMask(wr);
  rdPinSet   = digitalPinToBitMask(rd);
  csPinUnset = ~csPinSet;
  cdPinUnset = ~cdPinSet;
  wrPinUnset = ~wrPinSet;
  rdPinUnset = ~rdPinSet;
  #ifdef __AVR__
    *csPort   |=  csPinSet; // Set all control bits to HIGH (idle)
    *cdPort   |=  cdPinSet; // Signals are ACTIVE LOW
    *wrPort   |=  wrPinSet;
    *rdPort   |=  rdPinSet;
  #endif
  #if defined(__SAM3X8E__)
    csPort->PIO_SODR  |=  csPinSet; // Set all control bits to HIGH (idle)
    cdPort->PIO_SODR  |=  cdPinSet; // Signals are ACTIVE LOW
    wrPort->PIO_SODR  |=  wrPinSet;
    rdPort->PIO_SODR  |=  rdPinSet;
  #endif
  pinMode(cs, OUTPUT);    // Enable outputs
  pinMode(cd, OUTPUT);
  pinMode(wr, OUTPUT);
  pinMode(rd, OUTPUT);
  if(reset) {
    digitalWrite(reset, HIGH);
    pinMode(reset, OUTPUT);
  }
#endif

  init();
}

// Constructor for shield (fixed LCD control lines)
ST7783::ST7783(void) : Adafruit_GFX(TFTWIDTH, TFTHEIGHT) {
  init();
}

// Initialization common to both shield & breakout configs
void ST7783::init(void) {

#ifdef USE_ADAFRUIT_SHIELD_PINOUT
  CS_IDLE; // Set all control bits to idle state
  WR_IDLE;
  RD_IDLE;
  CD_DATA;
  digitalWrite(5, HIGH); // Reset line
  pinMode(A3, OUTPUT);   // Enable outputs
  pinMode(A2, OUTPUT);
  pinMode(A1, OUTPUT);
  pinMode(A0, OUTPUT);
  pinMode( 5, OUTPUT);
#endif

  setWriteDir(); // Set up LCD data port(s) for WRITE operations

  rotation  = 0;
  cursor_y  = cursor_x = 0;
  textsize  = 1;
  textcolor = 0xFFFF;
  _width    = TFTWIDTH;
  _height   = TFTHEIGHT;
  mirror_x = 0; 
  mirror_y = 0;
}

#define TFTLCD_DELAY 0xFF
 
static const uint16_t ST7781_regValues[] PROGMEM = {
  0x0001,0x0100,    
  0x0002,0x0700,    
  0x0003,0x1030,    
  0x0008,0x0302,    
  0x0009,0x0000,   
  0x000A,0x0008,    
  //*******POWER CONTROL REGISTER INITIAL*******//    
  0x0010,0x0790,    
  0x0011,0x0005,    
  0x0012,0x0000,   
  0x0013,0x0000,    
   //delayms(50, 
  //********POWER SUPPPLY STARTUP 1 SETTING*******//    
  0x0010,0x12B0,    
  // delayms(50,  
  0x0011,0x0007,    
   //delayms(50,  
  //********POWER SUPPLY STARTUP 2 SETTING******//    
  0x0012,0x008C,    
  0x0013,0x1700,    
  0x0029,0x0022,    
  // delayms(50,   
  //******GAMMA CLUSTER SETTING******//    
  0x0030,0x0000,    
  0x0031,0x0505,    
  0x0032,0x0205,    
  0x0035,0x0206,    
  0x0036,0x0408,    
  0x0037,0x0000,    
  0x0038,0x0504,
  0x0039,0x0206,    
  0x003C,0x0206,    
  0x003D,0x0408,    
  // -----------DISPLAY WINDOWS 240*320-------------//    
  0x0050,0x0000,
  0x0051,0x00EF,   
  0x0052,0x0000,   
  0x0053,0x013F,   
  //-----FRAME RATE SETTING-------//    
  0x0060,0xA700,   
  0x0061,0x0001,   
  0x0090,0x0033, //RTNI setting
  //-------DISPLAY ON------//    
  0x0007,0x0133,    0x0001,0x0100,    
   0x0002,0x0700,    
  0x0003,0x1030,    
  0x0008,0x0302,    
  0x0009,0x0000,   
  0x000A,0x0008,    
  //*******POWER CONTROL REGISTER INITIAL*******//    
  0x0010,0x0790,    
  0x0011,0x0005,    
  0x0012,0x0000,   
  0x0013,0x0000,    
   //delayms(50, 
  //********POWER SUPPPLY STARTUP 1 SETTING*******//    
  0x0010,0x12B0,    
  // delayms(50,  
  0x0011,0x0007,    
  // delayms(50,  
  //********POWER SUPPLY STARTUP 2 SETTING******//    
  0x0012,0x008C,    
  0x0013,0x1700,    
  0x0029,0x0022,    
  // delayms(50,   
  //******GAMMA CLUSTER SETTING******//    
  0x0030,0x0000,    
  0x0031,0x0505,    
  0x0032,0x0205,    
  0x0035,0x0206,    
  0x0036,0x0408,    
  0x0037,0x0000,    
  0x0038,0x0504,
  0x0039,0x0206,    
  0x003C,0x0206,    
  0x003D,0x0408,    
  // -----------DISPLAY WINDOWS 240*320-------------//    
  0x0050,0x0000,
  0x0051,0x00EF,   
  0x0052,0x0000,   
  0x0053,0x013F,   
  //-----FRAME RATE SETTING-------//    
  0x0060,0xA700,   
  0x0061,0x0001,   
  0x0090,0x0033, //RTNI setting
  //-------DISPLAY ON------//    
  0x0007,0x0133,   
};

void ST7783::begin() {

  uint16_t id = 0x9488;
  uint8_t i = 0;

  reset();

  delay(200);
	
  uint16_t a, d;
  CS_ACTIVE;
  while(i < sizeof(ST7781_regValues) / sizeof(uint16_t)) {
    a = pgm_read_word(&ST7781_regValues[i++]);
    d = pgm_read_word(&ST7781_regValues[i++]);
    if(a == TFTLCD_DELAY) delay(d);
    else                  writeRegister16(a, d);
  }
	setRotation (0);
  setAddrWindow(0, 0, TFTWIDTH-1, TFTHEIGHT-1);
	
	
  return;	

}

void ST7783::reset(void) {

  CS_IDLE;
//  CD_DATA;
  WR_IDLE;
  RD_IDLE;

#ifdef USE_ADAFRUIT_SHIELD_PINOUT
  digitalWrite(5, LOW);
  delay(2);
  digitalWrite(5, HIGH);
#else
  if(_reset) {
    digitalWrite(_reset, LOW);
    delay(2);
    digitalWrite(_reset, HIGH);
  }
#endif

  // Data transfer sync
  CS_ACTIVE;
  CD_COMMAND;
  write8(0x00);
  for(uint8_t i=0; i<3; i++) WR_STROBE; // Three extra 0x00s
  CS_IDLE;
}

void ST7783::setAddrWindow(int x1, int y1, int x2, int y2) {
  CS_ACTIVE;
  int x, y, t;
  switch(rotation) {
   default:
    x  = x1;
    y  = y1;
    break;
   case 1:
    t  = y1;
    y1 = x1;
    x1 = TFTWIDTH  - 1 - y2;
    y2 = x2;
    x2 = TFTWIDTH  - 1 - t;
    x  = x2;
    y  = y1;
    break;
   case 2:
    t  = x1;
    x1 = TFTWIDTH  - 1 - x2;
    x2 = TFTWIDTH  - 1 - t;
    t  = y1;
    y1 = TFTHEIGHT - 1 - y2;
    y2 = TFTHEIGHT - 1 - t;
    x  = x2;
    y  = y2;
    break;
   case 3:
    t  = x1;
    x1 = y1;
    y1 = TFTHEIGHT - 1 - x2;
    x2 = y2;
    y2 = TFTHEIGHT - 1 - t;
    x  = x1;
    y  = y2;
    break;
  }
    
  writeRegister16(0x0050, x1); // Set address window
  writeRegister16(0x0051, x2);
  writeRegister16(0x0052, y1);
  writeRegister16(0x0053, y2);
  writeRegister16(0x0020, x ); // Set address counter to top left
  writeRegister16(0x0021, y );
  
  CS_IDLE;
}



void ST7783::flood(uint16_t color, uint32_t len) {
  uint16_t blocks;
  uint8_t  i, hi = color >> 8,
              lo = color;

  CS_ACTIVE;
  CD_COMMAND;
  write8(0x00); // High byte of GRAM register...
  write8(0x22); // Write data to GRAM

  // Write first pixel normally, decrement counter by 1
  CD_DATA;
  write8(hi);
  write8(lo);
  len--;

  blocks = (uint16_t)(len / 64); // 64 pixels/block
  if(hi == lo) {
    // High and low bytes are identical.  Leave prior data
    // on the port(s) and just toggle the write strobe.
    while(blocks--) {
      i = 16; // 64 pixels/block / 4 pixels/pass
      do {
        WR_STROBE; WR_STROBE; WR_STROBE; WR_STROBE; // 2 bytes/pixel
        WR_STROBE; WR_STROBE; WR_STROBE; WR_STROBE; // x 4 pixels
      } while(--i);
    }
    // Fill any remaining pixels (1 to 64)
    for(i = (uint8_t)len & 63; i--; ) {
      WR_STROBE;
      WR_STROBE;
    }
  } else {
    while(blocks--) {
      i = 16; // 64 pixels/block / 4 pixels/pass
      do {
        write8(hi); write8(lo); write8(hi); write8(lo);
        write8(hi); write8(lo); write8(hi); write8(lo);
      } while(--i);
    }
    for(i = (uint8_t)len & 63; i--; ) {
      write8(hi);
      write8(lo);
    }
  }
  CS_IDLE;}

void ST7783::drawFastHLine(int16_t x, int16_t y, int16_t length, uint16_t color)
{

  int16_t x2;

  // Initial off-screen clipping
  if((length <= 0     ) ||
     (y      <  0     ) || ( y                  >= _height) ||
     (x      >= _width) || ((x2 = (x+length-1)) <  0      )) return;

  if(x < 0) {        // Clip left
    length += x;
    x       = 0;
  }
  if(x2 >= _width) { // Clip right
    x2      = _width - 1;
    length  = x2 - x + 1;
  }

  setAddrWindow(x, y, x2, y);
  flood(color, length);

}

void ST7783::drawFastVLine(int16_t x, int16_t y, int16_t length,
  uint16_t color)
{
  int16_t y2;

  // Initial off-screen clipping
  if((length <= 0      ) ||
     (x      <  0      ) || ( x                  >= _width) ||
     (y      >= _height) || ((y2 = (y+length-1)) <  0     )) return;
  if(y < 0) {         // Clip top
    length += y;
    y       = 0;
  }
  if(y2 >= _height) { // Clip bottom
    y2      = _height - 1;
    length  = y2 - y + 1;
  }

  setAddrWindow(x, y, x, y2);
  flood(color, length);

}

void ST7783::fillRect(int16_t x1, int16_t y1, int16_t w, int16_t h, 
  uint16_t fillcolor) {
  int16_t  x2, y2;

  // Initial off-screen clipping
  if( (w            <= 0     ) ||  (h             <= 0      ) ||
      (x1           >= _width) ||  (y1            >= _height) ||
     ((x2 = x1+w-1) <  0     ) || ((y2  = y1+h-1) <  0      )) return;
  if(x1 < 0) { // Clip left
    w += x1;
    x1 = 0;
  }
  if(y1 < 0) { // Clip top
    h += y1;
    y1 = 0;
  }
  if(x2 >= _width) { // Clip right
    x2 = _width - 1;
    w  = x2 - x1 + 1;
  }
  if(y2 >= _height) { // Clip bottom
    y2 = _height - 1;
    h  = y2 - y1 + 1;
  }

  setAddrWindow(x1, y1, x2, y2);
  flood(fillcolor, (uint32_t)w * (uint32_t)h);

}

void ST7783::fillScreen(uint16_t color) {  
  setAddrWindow(0, 0, _width - 1, _height - 1);
  flood(color, (long)TFTWIDTH * (long)TFTHEIGHT);
}

void ST7783::drawPixel(int16_t x, int16_t y, uint16_t color) {

  // Clip
  if((x < 0) || (y < 0) || (x >= _width) || (y >= _height)) return;

  CS_ACTIVE;
 
  int16_t t;
  switch(rotation) {
   case 1:
    t = x;
    x = TFTWIDTH  - 1 - y;
    y = t;
    break;
   case 2:
    x = TFTWIDTH  - 1 - x;
    y = TFTHEIGHT - 1 - y;
    break;
   case 3:
    t = x;
    x = y;
    y = TFTHEIGHT - 1 - t;
    break;
  }
  writeRegister16(0x0020, x);
  writeRegister16(0x0021, y);
  writeRegister16(0x0022, color );
	
  
  CS_IDLE;
}

void ST7783::pushColors(uint16_t *data, uint8_t len, boolean first) {
  uint16_t color;
  uint8_t  hi, lo;
  CS_ACTIVE;
  if(first == true) { // Issue GRAM write command only on first call
    CD_COMMAND;
    write8(0x00);
    write8(0x22);
  }
  CD_DATA;
  while(len--) {
    color = *data++;
    hi    = color >> 8; // Don't simplify or merge these
    lo    = color;      // lines, there's macro shenanigans
    write8(hi);         // going on.
    write8(lo);
  }
  CS_IDLE;
}

void ST7783::setRotation(uint8_t x) {

  // Call parent rotation func first -- sets up rotation flags, etc.
  Adafruit_GFX::setRotation(x);

}

#ifdef read8isFunctionalized
  #define read8(x) x=read8fn()
#endif

uint16_t ST7783::readPixel(int16_t x, int16_t y) {
  return 0;
}

// Ditto with the read/write port directions, as above.
uint16_t ST7783::readID(void) {

  uint16_t id = readReg(0x00);
  return id;
    
}

uint32_t ST7783::readReg(uint8_t r) {
  uint32_t id;
  uint8_t x;

  // try reading register #4
  CS_ACTIVE;
  CD_COMMAND;
  write8(r);
  setReadDir();  // Set up LCD data port(s) for READ operations
  CD_DATA;
  delayMicroseconds(50);

  read8(x);
  id = x;          // Do not merge or otherwise simplify
  id <<= 8;              // these lines.  It's an unfortunate
  read8(x);
  id  |= x;        // shenanigans that are going on.
  id <<= 8;              // these lines.  It's an unfortunate
  read8(x);
  id  |= x;        // shenanigans that are going on.
  id <<= 8;              // these lines.  It's an unfortunate
  read8(x);
  id  |= x;        // shenanigans that are going on.
  if (r == 0xEF){
    id <<= 8;         // these lines.  It's an unfortunate
	read8(x);
	id  |= x;	  
  }
  CS_IDLE;
  setWriteDir();  // Restore LCD data port(s) to WRITE configuration
  delay(150); //stabilization time
  return id;
}

// Pass 8-bit (each) R,G,B, get back 16-bit packed color
uint16_t ST7783::color565(uint8_t r, uint8_t g, uint8_t b) {
  return ((r & 0xF8) << 8) | ((g & 0xFC) << 3) | (b >> 3);
}

// For I/O macros that were left undefined, declare function
// versions that reference the inline macros just once:

#ifndef write8
void ST7783::write8(uint8_t value) {
  write8inline(value);
}
#endif

#ifdef read8isFunctionalized
uint8_t ST7783::read8fn(void) {
  uint8_t result;
  read8inline(result);
  return result;
}
#endif

#ifndef setWriteDir
void ST7783::setWriteDir(void) {
  setWriteDirInline();
}
#endif

#ifndef setReadDir
void ST7783::setReadDir(void) {
  setReadDirInline();
}
#endif

#ifndef writeRegister8
void ST7783::writeRegister8(uint8_t a, uint8_t d) {
  writeRegister8inline(a, d);
}
#endif

#ifndef writeRegister16

void ST7783::writeRegister16(uint16_t a, uint16_t d) {
  writeRegister16inline(a, d);
}
#endif

#ifndef writeRegisterPair
void ST7783::writeRegisterPair(uint8_t aH, uint8_t aL, uint16_t d) {
  writeRegisterPairInline(aH, aL, d);
}
#endif


void ST7783::writeRegister24(uint8_t r, uint32_t d) {
  CS_ACTIVE;
  CD_COMMAND;
  write8(r);
  CD_DATA;
  delayMicroseconds(10);
  write8(d >> 16);
  delayMicroseconds(10);
  write8(d >> 8);
  delayMicroseconds(10);
  write8(d);
  CS_IDLE;

}


void ST7783::writeRegister32(uint8_t r, uint32_t d) {
  CS_ACTIVE;
  CD_COMMAND;
  write8(r);
  CD_DATA;
  delayMicroseconds(10);
  write8(d >> 24);
  delayMicroseconds(10);
  write8(d >> 16);
  delayMicroseconds(10);
  write8(d >> 8);
  delayMicroseconds(10);
  write8(d);
  CS_IDLE;

}
